-- ************************************************************************** --
-- *************             BittWare Incorporated              ************* --
-- *************         9 Hills Ave, Concord, NH 03301         ************* --
-- ************************************************************************** --
-- LEGAL NOTICE:                                                              --
--                   Copyright (c) 2012 BittWare, Inc.                        --
--     The user is hereby granted a non-exclusive license to use and or       --
--       modify this code provided that it runs on BittWare hardware.         --
--     Usage of this code on non-BittWare hardware without the express        --
--        written permission of BittWare is strictly prohibited.              --
--                                                                            --
--  E-mail: support@bittware.com                    Tel: 603-226-0404         --
-- ************************************************************************** --

--------------------------------------------------------------------------------
-- Libraries
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.UtilityPkg.all;
use work.BoardInfoPkg.all;
use work.ComponentPkg.all;
use work.AvalonPkg.all;
use work.MiscPkg.all;
use work.RegisterPkg.all;
use work.FIFOPkg.all;
use work.ExamplePkg.all;

--------------------------------------------------------------------------------
-- ABSTRACT:
--------------------------------------------------------------------------------
--! @brief
--! The ExampleComponent is a fully functioning component that is available to
--! help users understand the basic AFW infrastructure features and how to implement
--! them. It explores and details two main areas, both in this document and in the
--! extensive use of comments in the source file. The first area is the use of the
--! RegisterBank component and its interface connections.  The second illustrates
--! the use of flow control FIFOs on an Avalon-ST sink input, and an Avalon-ST source output.

--------------------------------------------------------------------------------
-- DESIGN DETAILS:
--------------------------------------------------------------------------------
--! @details
--! It is worth repeating in this section, the main purpose of the ExampleComponent
--! is to familiarize the AFW user with the most widely used components in the architecture:
--! The RegisterBank and the flow control FIFO used on the Avalon-ST sink and Avalon-ST
--! source interface.  That being said, by its numerous comments and the coding itself,
--! the ExampleComponent illustrates several other features of AFW.  For example,
--! some functions from UtilityPkg file are used (toInt and resize).  The UtilityPkg
--! contains a multitude of functions that allow the re-use or simplification of some
--! commonly used code.  Another useful feature of the AFW illustrated in the
--! ExampleComponent is the coding style itself.  Constants are used to allow easy
--! scaling or relocation of resources with minimal effort.  A consistent naming
--! convention and careful formatting are used to make the code more easily readable.
--! Because of the depth of information found in this component, it is suggested
--! that the user read through the code fairly thoroughly to get a better sense
--! of some of the practical features of the AFW.

--------------------------------------------------------------------------------
-- NOTES:
--------------------------------------------------------------------------------
--! @note None

--------------------------------------------------------------------------------
-- TODO:
--------------------------------------------------------------------------------
--! @todo None

--------------------------------------------------------------------------------
-- BUGS
--------------------------------------------------------------------------------
--! @bug None

--------------------------------------------------------------------------------
-- RETURN LINKS:
--------------------------------------------------------------------------------
--! <br> @ref ExampleComponent "Return to Top"

--------------------------------------------------------------------------------
entity ExampleComponent is
--------------------------------------------------------------------------------
generic (
  --! @brief Board Specific Parameters
  BOARD_INFO    : BoardInfo;
  --! @brief Streaming Input Interface Parameters
  DATA_IN_INFO  : AvalonSTInfo;
  --! @brief Streaming Output Interface Parameters
  DATA_OUT_INFO : AvalonSTInfo
);

  --
  -- Other common generic interace parameters
  --

  -------------------------------
  -- Memory Interface Parameters
  -- The MASTER/SLAVE_INFO parameter is not used in most components.
  -- In the case of this example, we are restricting the address and data
  -- width of the memory interface so that we don't have to support reshaping of
  -- data and variable addres widths.  However, these parameters are still very
  -- necessary when you have a component that is decoding or switching
  -- the memory interface.
  -------------------------------
  -- Master Interface
  --MASTER_INFO  : AvalonMMMasterInfo
  -- Master Array Interface
  --MASTER_PORTS : natural;
  --MASTER_INFO  : AvalonMMMasterInfoArray;
  -------------------------------
  -- Slave Interface
  --SLAVE_INFO  : AvalonMMSlaveInfo;
  -- Slave Array Interface
  --SLAVE_PORTS : natural;
  --SLAVE_INFO  : AvalonMMSlaveInfoArray;
  -------------------------------
  -- In the rare occation when there is an input and output interface and both
  -- share the same configuration info, use the IN/OUT subfix to specify parameters.
  -- SLAVE/MASTER_IN/OUT_PORTS : natural;
  -- SLAVE/MASTER_IN/OUT_INFO  : AvalonMM(Master/Slave)Info;
  -- SLAVE/MASTER_IN/OUT_INFO  : AvalonMM(Master/Slave)InfoArray;
  -------------------------------
  -- Streaming Data Interface
  --DATA_INFO  : AvalonSTInfo;
  -- OR if the inputs and outputs don't match
  --DATA_IN_INFO  : AvalonSTInfo;
  --DATA_OUT_INFO : AvalonSTInfo;
  -------------------------------
  -- Streaming Array Interface
  --DATA_PORTS : natural;
  --DATA_INFO  : AvalonSTInfoArray;
  -- OR
  --DATA_INFO  : AvalonSTInfoArray(0 to DATA_PORTS-1);
  -------------------------------
  -- Streaming Array Input/Output Interfaces
  --DATA_IN_PORTS  : natural;
  --DATA_IN_INFO   : AvalonSTInfoArray;
  --DATA_OUT_PORTS : natural;
  --DATA_OUT_INFO  : AvalonSTInfoArray

port (
  --! @name Memory Mapped Slave Control/Status Interface
  --! @{
  slaveClk   : in  std_logic;
  slaveReset : in  std_logic;
  slaveIRQ   : out AvalonIRQSenderIntfArray(2 downto 0);
  slaveIn    : in  AvalonMMFabricToSlaveIntf;
  slaveInRsp : out AvalonMMSlaveToFabricIntf;
  --! @}
  --! @name Streaming Data Interfaces
  --! @{
  dataClk    : in  std_logic;
  dataReset  : in  std_logic;
  dataIn     : in  AvalonSTSourceToSinkIntf;
  dataInRsp  : out AvalonSTSinkToSourceIntf;
  dataOut    : out AvalonSTSourceToSinkIntf;
  dataOutRsp : in  AvalonSTSinkToSourceIntf
  --! @}
);

  --
  -- Other common Avalon Interface Combinations
  --

  -------------------------------
  -- Master Interface
  --masterClk    : in  std_logic;
  --masterReset  : in  std_logic;
  --masterIRQ    : in  AvalonIRQReceiverIntf;
  -- Master Input Interface
  --masterIn     : in  AvalonMMMasterToFabricIntf;
  --masterInRsp  : out AvalonMMFabricToMasterIntf;
  -- Master Output Interface
  --masterOut    : out AvalonMMMasterToFabricIntf;
  --masterOutRsp : in  AvalonMMFabricToMasterIntf;
  -------------------------------
  -- Slave Interface
  --slaveClk   : in  std_logic;
  --slaveReset : in  std_logic;
  --slaveIRQ   : out AvalonIRQSenderIntf;
  -- Slave Input Interface
  --slaveIn    : in  AvalonMMFabricToSlaveIntf;
  --slaveInRsp : out AvalonMMSlaveToFabricIntf;
  -- Slave Output Interface
  --slaveOut    : out AvalonMMFabricToSlaveIntf;
  --slaveOutRsp : in  AvalonMMSlaveToFabricIntf;
  -------------------------------
  -- Master Array Input Interface
  --masterIn     : in  AvalonMMMasterToFabricIntfArray(0 to MASTER_PORTS-1);
  --masterInRsp  : out AvalonMMFabricToMasterIntfArray(0 to MASTER_PORTS-1);
  -- Master Array Output Interface
  --masterOut    : out AvalonMMMasterToFabricIntfArray(0 to MASTER_PORTS-1);
  --masterOutRsp : in  AvalonMMFabricToMasterIntfArray(0 to MASTER_PORTS-1);
  -------------------------------
  -- Slave Array Input Interface
  --slaveIn     : in  AvalonMMFabricToSlaveIntfArray(0 to MASTER_PORTS-1);
  --slaveInRsp  : out AvalonMMSlaveToFabricIntfArray(0 to MASTER_PORTS-1);
  -- Slave Array Output Interface
  --slaveOut    : out AvalonMMFabricToSlaveIntfArray(0 to MASTER_PORTS-1);
  --slaveOutRsp : in  AvalonMMSlaveToFabricIntfArray(0 to MASTER_PORTS-1);
  -------------------------------
  -- Single Rate Component Clock Input
  --dataClk   : in std_logic;
  --dataReset : in std_logic;
  -- Multi-Rate Component Clock Input
  --dataInClk   : in std_logic;
  --dataInReset : in std_logic;
  -- Multi-Rate Component Clock Output
  --dataOutClk   : in std_logic;
  --dataOutReset : in std_logic;
  -------------------------------
  -- Streaming Input Interface
  --dataIn    : in AvalonSTSourceToSinkIntf;
  --dataInRsp : out AvalonSTSinkToSourcercIntf;
  -- Streaming Array Input Interface
  --dataIn    : in AvalonSTSourceToSinkIntfArray(0 to DATA_IN_PORTS-1);
  --dataInRsp : out AvalonSTSinkToSourceIntfArray(0 to DATA_IN_PORTS-1);
  -- Here is another example except you can defer the sizing.
  --dataIn    : in AvalonSTSourceToSinkIntfArray;
  --dataInRsp : out AvalonSTSinkToSourceIntfArray;
  -------------------------------
  -- Streaming Output Interface
  --dataOut    : out AvalonSTSourceToSinkIntf;
  --dataOutRsp : in AvalonSTSinkToSourceIntf;
  -- Streaming Array Output Interface
  --dataOut    : out AvalonSTSourceToSinkIntfArray(0 to DATA_OUT_PORTS-1);
  --dataOutRsp : in AvalonSTSinkToSourceIntfArray(0 to DATA_OUT_PORTS-1);
  -------------------------------
  -- Event Interface
  --eventCtrl   : out AvalonSTEventIntf;
  --eventStatus : in AvalonSTEventIntf;
  -- Event Interface Array
  --eventCtrl   : out AvalonSTEventIntfArray;
  --eventStatus : in AvalonSTEventIntfArray;

--------------------------------------------------------------------------------
end entity;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- BRIEF LINKS:
--------------------------------------------------------------------------------
--! @brief
--! <b>Entity Description:</b> @ref ExampleComponent "Here..."<br>
--! <b>Register Definitions:</b> @ref ExampleComponent_REGISTERS "Here..."<br>
--! <b>IRQ Definitions:</b> @ref ExampleComponent_IRQs "Here..."<br>
--! <b>Block Diagram:</b> @ref ExampleComponent_IMAGE "Here..."<br>
--! @details @anchor ExampleComponent_IMAGE

--------------------------------------------------------------------------------
-- BLOCK DIAGRAM:
--------------------------------------------------------------------------------
--! @details
--! The example component contains a RegisterBank, an input FIFO, and and output FIFO.
--! @image html example/docs/images/example_component.gif "Component Block Diagram"

--------------------------------------------------------------------------------
-- RETURN LINKS:
--------------------------------------------------------------------------------
--! <br> @ref ExampleComponent_BEH "Return to Top"

--------------------------------------------------------------------------------
architecture ExampleComponent_BEH of ExampleComponent is
--------------------------------------------------------------------------------

  ---------------------------
  -- Size Declarations
  ---------------------------

  -- By declaring some range types, access of data is easier.  Add more, remove
  -- some.  Do whatever makes it easiest to get at the data fields and meets
  -- design requirements.  Oh, just don't hardcode BITS
  -- For ports using a single sizing...use DATA_INFO/WIDTH/RANGE
  --constant DATA_WIDTH : natural := DATA_INFO.BITS_PER_SYMBOL*DATA_INFO.SYMBOLS_PER_BEAT;
  --subtype  DATA_RANGE is natural range (DATA_WIDTH-1) downto 0;

  --! @brief Avalon-ST data in width
  constant DATA_IN_WIDTH : natural := DATA_IN_INFO.BITS_PER_SYMBOL*DATA_IN_INFO.SYMBOLS_PER_BEAT;
  subtype  DATA_IN_RANGE is natural range (DATA_IN_WIDTH-1) downto 0;

  --! @brief Avalon-ST data out width
  constant DATA_OUT_WIDTH : natural := DATA_OUT_INFO.BITS_PER_SYMBOL*DATA_OUT_INFO.SYMBOLS_PER_BEAT;
  subtype  DATA_OUT_RANGE is natural range (DATA_OUT_WIDTH-1) downto 0;

  ---------------------------
  -- Register Configuration
  ---------------------------

  --! @brief Captures the default register bit width for registers not that dont have specifically defined bits.
  constant REG_DATA_WIDTH : natural := 8;
  subtype  REG_DATA_RANGE is natural range (REG_DATA_WIDTH-1) downto 0;

  ------------------------------------------------------------------------------
  -- DOXYGEN INSTRUCTIONS
  ------------------------------------------------------------------------------
  -- DOXYGEN: Register Description Instructions:
  -- DOXYGEN: Add a group named for this component.  All of the component's registers
  -- DOXYGEN: will be add to this group. Describe the group detail with the
  -- DOXYGEN: tag, @details, and add an @see to link back to the architecture.
  -- DOXYGEN: Create a group with the same name as the register.  Describe the
  -- DOXYGEN: register with @details tag and add an @see link back to the architecture
  -- DOXYGEN: The @{ and @} bound the group.  The entity group name must be unique.
  -- DOXYGEN: Describe the register bits with the @details tag
  ------------------------------------------------------------------------------

  --! @defgroup ExampleComponent_REGISTERS ExampleComponent Registers
  --! @details
  --! None of the example component registers are optional.
  --! @see ExampleComponent
  --! @see ExampleComponent_BEH

  --! @defgroup LCR CONTROL_REGISTER
  --! @ingroup ExampleComponent_REGISTERS
  --! @details Example custom local control register.
  --! @{
  constant CONTROL_REGISTER : RegisterLocation := init( "example_control", REG_COMPONENT_START_ADDR+toInt(x"00"), 3, RW_REG_CFG );
  --! @details Custom single bit in the control register.
  constant A_CTRL_BIT : RegisterBit := 0;
  --! @details This range defines a field within the register and can
  --! be used for access.
  subtype SOME_CTRL_RANGE is natural range 2 downto 1;
  --! @}

  --! @defgroup LSR STATUS_REGISTER
  --! @ingroup ExampleComponent_REGISTERS
  --! @details Example custom loacal status register. Reuse FIFOPkg
  --! bit defines to promote consistancy.  All status bits are tied
  --! to the input FIFO.  In many cases you will not want to reuse
  --! the FIFO bits, that is up to you as every component may have
  --! more than one FIFO.
  --! @{
  constant STATUS_REGISTER : RegisterLocation := init( "exmaple_status", REG_COMPONENT_START_ADDR+toInt(x"01"), 8, RO_STATUS_REG_CFG );
  --! @details Input FIFO write side is almost full
  constant ALMOST_FULL_BIT  : RegisterBit := FIFO_WR_ALMOST_FULL_BIT;
  --! @details Input FIFO write side is full
  constant FULL_BIT         : RegisterBit := FIFO_WR_FULL_BIT;
  --! @details Input FIFO read side is almost empty
  constant ALMOST_EMPTY_BIT : RegisterBit := FIFO_RD_ALMOST_EMPTY_BIT;
  --! @details Input FIFO read side is empty
  constant EMPTY_BIT        : RegisterBit := FIFO_RD_EMPTY_BIT;
  --! @}

  --! @defgroup CCR COUNT_CONFIG_REGISTER
  --! @ingroup ExampleComponent_REGISTERS
  --! @details Used to show an example write clear configuration.
  --! @{
  constant COUNT_CONFIG_REGISTER : RegisterLocation := init( "count_config", REG_COMPONENT_START_ADDR+toInt(x"02"), 2, RW_WR_CLEAR_REG_CFG );
  --! @details Load the counter value from @ref CIR
  constant LOAD_BIT  : RegisterBit := 0;
  --! @details Pause the counter and preserve the value
  constant PAUSE_BIT : RegisterBit := 1;
  --! @}

  --! @defgroup CIR COUNT_INIT_REGISTER
  --! @ingroup ExampleComponent_REGISTERS
  --! @details This is an example standard read/write register.  Write to this
  --! register to set the internal counter.  Value is latched in when the
  --! @ref LOAD_BIT is set.  Width is defined by @ref REG_DATA_WIDTH.
  --! @{
  constant COUNT_INIT_REGISTER : RegisterLocation := init( "count_init", REG_COMPONENT_START_ADDR+toInt(x"03"), REG_DATA_WIDTH, RW_REG_CFG );
  --! @}

  --! @defgroup CSR COUNT_STEP_REGISTER
  --! @ingroup ExampleComponent_REGISTERS
  --! @details Write this register to change the step when the counter updates.
  --! The counter will count down by this amount.  Width is defined by @ref REG_DATA_WIDTH
  --! @{
  constant COUNT_STEP_REGISTER : RegisterLocation := init( "count_step", REG_COMPONENT_START_ADDR+toInt(x"04"), REG_DATA_WIDTH, RW_REG_CFG );
  --! @}

  --! @defgroup FR FLAGS_REGISTER
  --! @ingroup ExampleComponent_REGISTERS
  --! @details Used to show an example read only configuration.  The whole register will be set
  --! to all ones when the counter expires.  After the register is read, the flags will clear.
  --! Width is defined by @ref REG_DATA_WIDTH
  --! @{
  constant FLAGS_REGISTER : RegisterLocation := init( "flags", REG_COMPONENT_START_ADDR+toInt(x"05"), REG_DATA_WIDTH, RO_STATUS_RD_FLAG_REG_CFG );
  --! @}

  --! @defgroup TIR TRIGGER_IRQ_REGISTER
  --! @ingroup ExampleComponent_REGISTERS
  --! @details This register location contains a level triggered IRQ
  --! and you will see the associated IRQ bit trigger, assuming the IRQ mask has been set.
  --! See the @ref RegisterPkg.RB_IRQ_STAT "IRQ Status Register" for more details
  --! on its operation.
  --! @{
  --! @details Trigger rising edge IRQ
  constant IRQ_BIT  : RegisterBit := 0;
  --! @details The number of IRQs per IRQ Status register
  constant NUM_IRQ_BITS : natural := IRQ_BIT + 1;

  constant TRIGGER_IRQ_REGISTER : RegisterLocation := init( "trigger_irq", REG_COMPONENT_START_ADDR+toInt(x"06"), NUM_IRQ_BITS, RW_REG_CFG );
  --! @}

  --! @defgroup LISR LEVEL_IRQ_STATUS_REGISTER
  --! @ingroup ExampleComponent_REGISTERS
  --! @details This register location is used to trigger the IRQs since the component has no
  --! existing internal IRQ trigger logic.  It will show how each different types of triggers work.
  --! This is not necessary to get IRQs working, it is just used here as the driver to set
  --! the IRQ bits defined below.  Simply write a '1' to the corresponding bit in this register
  --! and you will see the associated IRQ bit trigger, assuming the IRQ mask has been set.
  --! @{
  constant LEVEL_IRQ_STATUS_REGISTER : RegisterLocation := init( "level_irq_status", toInt(x"08"), NUM_IRQ_BITS, LEVEL_IRQ_REG_CFG );
  --! @}

  --! @defgroup LIMR LEVEL_IRQ_MASK_REGISTER
  --! @ingroup ExampleComponent_REGISTERS
  --! @details This creates a IRQ Mask register for the @ref LISR LEVEL_IRQ_STATUS_REGISTER at
  --! the previous address see the @ref RegisterPkg.RB_IRQ_MASK "IRQ Mask Register" for more
  --! details on its operation
  --! be unmasked.
  --! @{
  constant LEVEL_IRQ_MASK_REGISTER : RegisterLocation := init( "level_irq_mask", toInt(x"09"), NUM_IRQ_BITS, STD_IRQ_MASK_REG_CFG );
  --! @}

  --! @defgroup FESR FALLING_EDGE_IRQ_STATUS_REGISTER
  --! @ingroup ExampleComponent_REGISTERS
  --! @details This creates a Falling Edge Triggered IRQ register.  Since there is no IRQ mask
  --! register at the next register location, by default all IRQs in this status register will
  --! be unmasked.  See the @ref RegisterPkg.RB_IRQ_STAT "IRQ Status Register" for more details
  --! on its operation.
  --! @{
  constant FALLING_EDGE_IRQ_STATUS_REGISTER : RegisterLocation := init( "falling_edge_irq_status", toInt(x"0A"), NUM_IRQ_BITS, FALLING_EDGE_IRQ_REG_CFG );
  --! @}

  --! @brief All of the custom register locations.
  constant COMPONENT_REG_LOCATIONS : RegisterLocationArray := ( CONTROL_REGISTER,
                                                                STATUS_REGISTER,
                                                                COUNT_CONFIG_REGISTER,
                                                                COUNT_INIT_REGISTER,
                                                                COUNT_STEP_REGISTER,
                                                                FLAGS_REGISTER,
                                                                TRIGGER_IRQ_REGISTER,
                                                                LEVEL_IRQ_STATUS_REGISTER,
                                                                LEVEL_IRQ_MASK_REGISTER,
                                                                FALLING_EDGE_IRQ_STATUS_REGISTER );

  -- If for some reason, only 1 RegisterLocation is desired, the syntax would be as follows:
  --constant COMPONENT_REG_LOCATIONS : RegisterLocationArray := (0 => CONTROL_REGISTER);

  ---------------------------
  -- Register Bank Setup
  ---------------------------

  -- NOTE: All the standard addresses are defined in RegisterPkg.vhd

  -- UPDATE THE COMPONENT VERSION NUMBER...
  --! @brief Version number of this component
  constant VERSION : natural := 3;

  --! @brief Definition of this component's registers.
  --! @details There are multiple types of RegisterBank supported.  Standard and small.
  --! A STANDARD_?_REG_BANK is used when a component is in stand alone mode and the memory
  --! interface needs access to the register values.  SMALL_REG_BANK is used when the
  --! register bank is in a sub-component and we only need minor status which will be tied
  --! to a parent register bank for communication over the memory interface.

  constant STD_REG_LOC_CFG : StandardRegisterLocationsConfig := init(ID_EXAMPLE_COMPONENT, VERSION, RISING_EDGE_IRQ, NUM_IRQ_BITS);

  constant REGISTER_BANK_INFO : RegisterBankInfo := init( STD_REG_LOC_CFG, COMPONENT_REG_LOCATIONS );

  ---------------------------
  -- FIFO Configuration
  ---------------------------

  --! @brief Depth used for both the input and output FIFO components
  constant FIFO_DEPTH : natural := 8;

  --! @brief Used to configure the input FIFO
  constant FIFO_IN_FIFO_INFO : FIFOInfo := ( FIFO_TYPE           => RAM_SINGLE_RATE_FIFO,
                                             DEPTH               => FIFO_DEPTH,
                                             WR_ALMOST_FULL_CNT  => FIFO_DEPTH-1,
                                             WR_ALMOST_EMPTY_CNT => FIFO_REMOVE_COUNT,
                                             RD_ALMOST_FULL_CNT  => FIFO_REMOVE_COUNT,
                                             RD_ALMOST_EMPTY_CNT => 1,
                                             SHOW_AHEAD          => true );

  --! @brief Used to configure the output FIFO
  constant FIFO_OUT_FIFO_INFO : FIFOInfo := init( RAM_SINGLE_RATE_FIFO, FIFO_DEPTH );

  ---------------------------
  -- Signal Declarations
  ---------------------------

  -- Reset Signals
  signal reset_in   : std_logic;
  signal data_reset : std_logic;

  -- Custom Registers
  signal control_bit   : std_logic;
  signal control_field : std_logic_vector(SOME_CTRL_RANGE);

  -- Write clear signals
  signal load  : std_logic;
  signal pause : std_logic;

  signal count_init : std_logic_vector(REG_DATA_RANGE);
  signal count_step : std_logic_vector(REG_DATA_RANGE);

  -- Read Clear Signals
  signal clear_flags : std_logic;
  signal flags       : std_logic_vector(REG_DATA_RANGE);

  -- IRQ example signal
  signal irq_flag  : std_logic;

  -- Register Bank Signals
  signal reg_ctrl   : RegisterControlIntf;
  signal reg_status : RegisterStatusIntf;

  -- Input FIFO signals
  signal fifo_in_status       : AvalonMMSlaveToFabricIntf;
  signal fifo_in_data_out     : AvalonSTSourceToSinkIntf;
  signal fifo_in_data_out_rsp : AvalonSTSinkToSourceIntf;

  -- Output FIFO signals
  signal fifo_out_data_in     : AvalonSTSourceToSinkIntf;
  signal fifo_out_data_in_rsp : AvalonSTSinkToSourceIntf;

  ---------------------------
  -- Dump XML
  ---------------------------

  -- synthesis translate_off
  use work.ctype_h.all;
  use work.strings_h.all;
  use work.stdio_h.all;

  impure function dump_xml return boolean is
    constant NAME      : string := ExampleComponent'path_name;
    constant FILE_NAME : FileName := init( "project_config.xml" );
    variable xml_file  : CFILE;
  begin
    xml_file := fopen(FILE_NAME, "a");

    if( xml_file = 0 ) then
      report "ERROR: ExampleComponent.vhd - Cannot Open XML File " &  FILE_NAME
        severity failure;
    end if;

    fprintf(xml_file, "<ComponentInfo type=%s name=%s>\n", addQuotes("ExampleComponent"), addQuotes(NAME) );
    fprintf(xml_file, "%s", getXML("data_in_info", DATA_IN_INFO));
    fprintf(xml_file, "%s", getXML("data_out_info", DATA_OUT_INFO));
    fprintf(xml_file, "</ComponentInfo>\n");

    fclose(xml_file);

    return true;
  end function;

  constant DUMP_XML_FILES : boolean := dump_xml;
  -- synthesis translate_on

--------------------------------------------------------------------------------
begin
--------------------------------------------------------------------------------

  ------------------------------------------------------------------------------
  -- Component Assertions
  ------------------------------------------------------------------------------

  -- Example Assert
  assert( true ) report "ExampleComponent.vhd - Message goes here..." severity error;

  assert( DATA_IN_INFO.PROFILE = AVALON_ST_TYPICAL )
    report "ExampleComponent.vhd - Input profile must be AVALON_ST_TYPICAL"
    severity failure;

  assert( DATA_OUT_INFO.PROFILE = AVALON_ST_TYPICAL )
    report "ExampleComponent.vhd - Output profile must be AVALON_ST_TYPICAL"
    severity failure;

  ------------------------------------------------------------------------------
  example_gen : if( false ) generate
  ------------------------------------------------------------------------------
    -- generate specific declarations here...
  begin

    -- This isn't needed for this example, but here is how you do one...

  ------------------------------------------------------------------------------
  end generate;
  ------------------------------------------------------------------------------

  ------------------------------------------------------------------------------
  -- Register Bank
  ------------------------------------------------------------------------------

  register_bank_map : RegisterBank
  generic map (
    BOARD_INFO         => BOARD_INFO,
    REGISTER_BANK_INFO => REGISTER_BANK_INFO
  )
  port map (
    -- Avalon-MM Slave Interfaces
    slaveClk    => slaveClk,
    slaveReset  => slaveReset,
    slaveIRQ    => slaveIRQ,
    slaveIn     => slaveIn,
    slaveInRsp  => slaveInRsp,
    -- Register Interface
    regClk      => dataClk,
    regReset    => dataReset,
    regCtrl     => reg_ctrl,
    regStatus   => reg_status
  );

  ------------------------------------------------------------------------------
  -- Register Soft Reset
  ------------------------------------------------------------------------------

  -- Please don't use the soft reset to reset the RegisterBank slave
  -- or register interface.  It won't work, I PROMISE!

  reset_in <= dataReset or reg_ctrl.reset;

  -- We don't want to drive reset lines with combinational logic
  -- In this case, resets are already synchronized, just register the signal.
  -- If we needed to synch to a different clock domain, this
  -- soft reset component will come in handy.
  soft_reset_map : SoftReset
  generic map (
    -- Min Cycles to Hold Reset
    DELAY => 1
  )
  port map (
    clk  => dataClk,
    arst => reset_in,
    srst => data_reset
  );

  ------------------------------------------------------------------------------
  -- Register Control
  ------------------------------------------------------------------------------

  -- NOTE:
  -- If data and mem clocks are not the same, this will perform a retime and
  -- ease up on timing constraints.  Anything critical requires a fifo
  -- for retiming. Especially on the readback side.
  -- The best method for retiming a control signal is to have the master
  -- drive all the configuration registers and then set a single bit
  -- to latch in the update using a the read and write pulse/clear register access
  -- modes. The standard register bank will automaticically retime those flags.

  --! @brief Capture RegisterBank control
  reg_ctrl_prc : process( dataClk )
  begin
    if( dataClk'event and dataClk = '1' ) then
      if( data_reset = '1' ) then
        -- If necessary, clear the control signals here...
        control_bit       <= '0';
        control_field     <= (others => '0');
        load              <= '0';
        pause             <= '0';
        count_init        <= (others => '0');
        count_step        <= (others => '0');
        clear_flags       <= '0';
        irq_flag  <= '0';
      else
        -- Some example read/write registers.  One could also use the
        -- getValue functions defined in RegisterPkg.vhd
        control_bit   <= reg_ctrl.memory(getAddr(CONTROL_REGISTER))(A_CTRL_BIT);
        control_field <= reg_ctrl.memory(CONTROL_REGISTER.ADDR)(control_field'range);
        -- OR: control_field <= reg_ctrl.memory(getAddr(CONTROL_REGISTER))(SOME_CTRL_RANGE);
        -- OR: control_field <= getValue( reg_ctrl, getAddr(CONTROL_REGISTER), control_field );
        -- OR: control_field <= getValue( reg_ctrl, getAddr(CONTROL_REGISTER), control_field'length, control_field'low );

        -- Next is a block of configuration registers that should only be updated once.
        -- For instace, with a write clear register.  In this case, we will allow
        -- the host to write all the configs and then set a load bit to latch in the new config.
        -- This can be done with a write pulse also, in which case the last register written
        -- to triggers a pulse, to indicate the data write is completed.
        -- With a write clear, the flag will be cleared so that it is only active for
        -- a single cycle.  Be sure to set the access mode to write clear.
        -- The write clear type register can also be used to latch status, increment
        -- a counter, write a fifo, or...
        -- The register bank will automatically retime the flags from the slave clock
        -- to the register interface clock in standard mode.
        load <= '0';
        if( getFlag(reg_ctrl, COUNT_CONFIG_REGISTER) = '1' ) then
          load       <= getValue( reg_ctrl, COUNT_CONFIG_REGISTER, LOAD_BIT );
          count_init <= getValue( reg_ctrl, COUNT_INIT_REGISTER, count_init );
          count_step <= getValue( reg_ctrl, COUNT_STEP_REGISTER, count_step );

          -- In some cases, you may want to latch a write clear register.  In that case,
          -- check the write clear flag and then grab your data.  Be careful doing this
          -- with a bus though because all the bits might not be stable crossing clock
          -- domains...if you aren't using the standard register bank.  Always best with
          -- a single bit.
          -- Latch the bit only when the register is written,
          pause <= getValue(reg_ctrl, COUNT_CONFIG_REGISTER, PAUSE_BIT);
        end if;

        -- Any of the read clear registers implemented should have an associated
        -- flag that gets set when the address is read.  The register bank
        -- feeds back a flag so that the component knows to perform the clear.
        --clear_flags <= reg_ctrl.flags(FLAGS_REGISTER.ADDR);
        -- OR
        clear_flags <= getFlag(reg_ctrl, FLAGS_REGISTER);
        -- OR
        -- One could even check another register when the flag is set and
        -- use that to determine if some flags should be cleared or not.
        --clear_flags <= getValue( reg_ctrl, CLEAR_CTRL_REGISTER, CLEAR_FLAGS_BIT );

        -- This register is used to set the IRQ bits as an example
        irq_flag  <= getValue( reg_ctrl, TRIGGER_IRQ_REGISTER, IRQ_BIT );

      end if;
    end if;
  end process;

  ------------------------------------------------------------------------------
  -- Register Status
  ------------------------------------------------------------------------------

  --! @brief Write back register status
  reg_stat_prc : process( dataClk )
  begin
    -- This process doesn't need to be clocked if slaveClk = dataClk but it won't do
    -- anything but add a cycle of latency over a slow status interface anyways.
    -- Don't forget to somehow clear the status on reset.
    if( dataClk'event and dataClk = '1' ) then
      if( data_reset = '1' ) then
        reg_status <= initRegisterStatusIntf;
      else
        -- Custom Component Registers
        reg_status.memory(STATUS_REGISTER.ADDR)(ALMOST_FULL_BIT)  <= fifo_in_status.readdata(FIFO_WR_ALMOST_FULL_BIT);
        reg_status.memory(STATUS_REGISTER.ADDR)(FULL_BIT)         <= fifo_in_status.readdata(FIFO_WR_FULL_BIT);
        reg_status.memory(STATUS_REGISTER.ADDR)(ALMOST_EMPTY_BIT) <= fifo_in_status.readdata(FIFO_RD_ALMOST_EMPTY_BIT);
        reg_status.memory(STATUS_REGISTER.ADDR)(EMPTY_BIT)        <= fifo_in_status.readdata(FIFO_RD_EMPTY_BIT);

        -- Below is an example read clear register.  Basically, each time a register of this type
        -- is read, the register bank will send us a flag letting us know.  When the flag is seen,
        -- that is a trigger to clear the internal component register.  The read clear can be
        -- used for flags, interrupts, or any other type of transaction with a desired read and
        -- clear. The control process is used to check the read clear flag.
        -- A similar register access mode is the read pulse, however, the register need not be
        -- cleared on the pulse.
        setValue( reg_status, FLAGS_REGISTER, flags );

        -- Set the IRQ status register bits using a procedure.
        setIRQ(reg_status, IRQ_BIT, irq_flag);
        setIRQ(reg_status, IRQ_BIT, irq_flag, LEVEL_IRQ_STATUS_REGISTER);
        setIRQ(reg_status, IRQ_BIT, irq_flag, FALLING_EDGE_IRQ_STATUS_REGISTER);
      end if;
    end if;
  end process;

  ------------------------------------------------------------------------------
  -- Counter Process
  ------------------------------------------------------------------------------

  reg_clr_prc : process( dataClk )
    variable cnt      : natural;
    variable done     : boolean;
    constant INIT_CNT : natural := 10;
  begin
    if( dataClk'event and dataClk = '1' ) then
      if( data_reset = '1' ) then
        cnt   := INIT_CNT;
        done  := true;
        flags <= (others => '0');
      elsif( pause = '0' ) then

        -- When we see the clear flag read event, clear the
        -- flags register.
        if( load = '1' or clear_flags = '1' ) then
          flags <= (others => '0');
        end if;

        -- Looks like we got a load event
        if( load = '1' ) then
          cnt  := toInt(count_init);
          done := false;
        end if;

        -- Set the flags when our counter is done to create some
        -- status.
        if( done = false and cnt = 0 ) then
          flags <= (others => '1');
          done  := true;
        end if;

        if( done = false ) then
          cnt := cnt - toInt(count_step);
        end if;

      end if;
    end if;
  end process;

  ------------------------------------------------------------------------------
  -- Input FIFO
  ------------------------------------------------------------------------------

  -- FIFO is used for flow control.  If a depth of more than 16 or so
  -- is required, you would go to a RAM based FIFO...
  -- The single rate fifo is register based so you don't always want to use it.
  -- HINT: using a fifo on the input makes avalon backpressure easier to implement
  -- without dropping data at certain times, so it is not recommeneded to
  -- take the fifo out.
  fifo_in_map : FIFO
  generic map (
    BOARD_INFO   => BOARD_INFO,
    REG_BANK_CFG => SMALL_REG_BANK_CONFIG,
    -- Avalon-ST Parameters
    DATA_INFO    => DATA_IN_INFO,
    -- FIFO Parameters
    FIFO_INFO    => FIFO_IN_FIFO_INFO
  )
  port map (
    -- Avalon-MM Slave Interfaces
    slaveInRsp   => fifo_in_status,
    -- Avalon-ST Input Interface
    dataInClk    => dataClk,
    dataInReset  => data_reset,
    dataIn       => dataIn,
    dataInRsp    => dataInRsp,
    -- Avalon-ST Output Interface
    dataOutClk   => dataClk,
    dataOutReset => data_reset,
    dataOut      => fifo_in_data_out,
    dataOutRsp   => fifo_in_data_out_rsp
  );

  ------------------------------------------------------------------------------
  -- Processing Code and Output Combined
  ------------------------------------------------------------------------------

  -- Apply back pressure when output is not ready
  fifo_in_data_out_rsp.ready <= fifo_out_data_in_rsp.ready
                                when (data_reset = '0' and reg_ctrl.enable = '1')
                                else '0';

  -- Let's do something here.
  -- Assume that SYMBOLS_PER_BEAT is equal to one to avoid a data reshape.
  -- In some components the data may need to be reshaped.  Usually in multi-rate components.
  -- Also assuming AVALON_ST_TYPICAL profile.
  --
  -- NOTE: This process was implemented specifically for communication with an output
  -- data FIFO.  Otherwise, this code would violate the Avalon spec.  It utilizes the
  -- response ready flag as a FIFO not empty flag.  Following the Avalon spec,
  -- it is required that the valid is NOT set conditional on the ready when the ready
  -- latency is set to zero. Valid should typically be independent of ready.
  do_prc : process( dataClk )
    variable disable_while_valid : boolean;
  begin
    if( dataClk'event and dataClk = '1' ) then
      if( data_reset = '1' ) then
        fifo_out_data_in    <= initAvalonSTSourceToSinkIntf;
        disable_while_valid := false;
      elsif( reg_ctrl.enable = '1' ) then

        -- At this point, a few conditions could have occured.
        -- 1. We set valid on the previous cycle and it was accepted.
        -- 2. We set valid on the previous cycle and it was not accepted.
        -- 3. We didn't send data at all on the previous cycle

        -- Check the disable flag to help us restart
        if( disable_while_valid = true ) then
          fifo_out_data_in.valid <= '1';
          disable_while_valid    := false;

        -- The sink was ready to accept data on the previous cycle
        elsif( fifo_out_data_in_rsp.ready = '1' ) then

          -- Condition #1
          -- Condition #3
          -- Either way, transfer the next data point.  This is where our
          -- signal processing would occur.
          fifo_out_data_in.valid                <= fifo_in_data_out.valid;
          fifo_out_data_in.data(DATA_OUT_RANGE) <= resize( fifo_in_data_out.data(DATA_IN_RANGE), DATA_OUT_WIDTH );

        -- The sink was not ready to accept data on the previous cycle
        else

          -- Condition #2
          if( fifo_out_data_in.valid = '1' ) then
            -- Hold right where we are
            fifo_out_data_in.valid <= '1';
          -- Condition #3
          else
            -- Don't do anything once again. But you can present
            -- the next data here if you want to stick to the
            -- Avalon spec.
            fifo_out_data_in.valid <= '0';
          end if;

        end if;

      else
        -- The component is disabled so stop all transfers
        fifo_out_data_in.valid <= '0';

        -- Set a disable flag if the data was not accepted
        if( fifo_out_data_in.valid = '1' and fifo_out_data_in_rsp.ready = '0' ) then
          disable_while_valid := true;
        end if;

      end if;

    end if;
  end process;

  ------------------------------------------------------------------------------
  -- Output FIFO
  ------------------------------------------------------------------------------

  fifo_out_map : FIFO
  generic map (
    BOARD_INFO => BOARD_INFO,
    -- Avalon-ST Parameters
    DATA_INFO  => DATA_OUT_INFO,
    -- FIFO Parameters
    FIFO_INFO  => FIFO_OUT_FIFO_INFO
  )
  port map (
    -- Avalon-ST Input Interface
    dataInClk    => dataClk,
    dataInReset  => data_reset,
    dataIn       => fifo_out_data_in,
    dataInRsp    => fifo_out_data_in_rsp,
    -- Avalon-ST Output Interface
    dataOutClk   => dataClk,
    dataOutReset => data_reset,
    dataOut      => dataOut,
    dataOutRsp   => dataOutRsp
  );

--------------------------------------------------------------------------------
end architecture;
--------------------------------------------------------------------------------
