-- ************************************************************************** --
-- *************             BittWare Incorporated              ************* --
-- *************         9 Hills Ave, Concord, NH 03301         ************* --
-- ************************************************************************** --
-- LEGAL NOTICE:                                                              --
--                   Copyright (c) 2011 BittWare, Inc.                        --
--     The user is hereby granted a non-exclusive license to use and or       --
--       modify this code provided that it runs on BittWare hardware.         --
--     Usage of this code on non-BittWare hardware without the express        --
--        written permission of BittWare is strictly prohibited.              --
--                                                                            --
--  E-mail: support@bittware.com                    Tel: 603-226-0404         --
-- ************************************************************************** --

--------------------------------------------------------------------------------
-- Libraries
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.UtilityPkg.all;
use work.BoardInfoPkg.all;
use work.AvalonPkg.all;
use work.DMAPkg.all;

--------------------------------------------------------------------------------
-- ABSTRACT:
--------------------------------------------------------------------------------
--! @brief
--! This package serves as an example package used by the ExampleComponent.

--------------------------------------------------------------------------------
-- DESIGN DETAILS:
--------------------------------------------------------------------------------
--! @details

--------------------------------------------------------------------------------
-- NOTES:
--------------------------------------------------------------------------------
--! @note None

--------------------------------------------------------------------------------
-- TODO:
--------------------------------------------------------------------------------
--! @todo None

--------------------------------------------------------------------------------
-- BUGS
--------------------------------------------------------------------------------
--! @bug None

--------------------------------------------------------------------------------
-- RETURN LINKS:
--------------------------------------------------------------------------------
--! <br> @ref ExamplePkg "Return to Top"

--------------------------------------------------------------------------------
package ExamplePkg is
--------------------------------------------------------------------------------

  ------------------------------------------------------------------------------
  -- ExampleComponent
  ------------------------------------------------------------------------------

  ---------------------------
  -- Constant Declarations
  ---------------------------

  --! @brief An example constant
  constant CONST_VALUE : integer := 128;

  --! @details bitWidth(AVALON_ST_MAX_DATA_WIDTH) is the largest it can ever be. (log2)
  --! That is assuming all but ONE 1 bit symbols are empty during the beat.
  constant MY_CONSTANT : integer := bitWidth(CONST_VALUE);

  ---------------------------
  -- Type Declarations
  ---------------------------

  --! @brief An example subtype
  subtype MY_DATA_RANGE is natural range 7 downto 0;

  --! @brief An example custom type
  type MyType is array (natural range <>) of std_logic;

  --! @brief An example custom type array
  type MyDataArray is array (natural range <>) of std_logic_vector(CONST_VALUE-1 downto 0);

  --! @brief An example record
  --! @details
  type MyRecord is record
    --! @brief This is a valid field
    valid : std_logic;
    --! @brief This is a data field
    data  : std_logic_vector(MY_DATA_RANGE);
  end record;

  --! @brief An array of records
  type ArrayOfMyRecord is array(natural range <>) of MyRecord;

  ---------------------------
  -- Function Declarations
  ---------------------------

  --! @brief An example function with no arguments.
  --! @details
  --! @returns Some value of type @ref MyType.
  function myFunction return MyType;

  --! @brief An example function with an argument list.
  --! @details
  --! @param[in] x Passed in @ref MyType
  --! @param[in] y Passed in first argument.
  --! @param[in] z Passed in second argument.
  --! @returns Some value of @ref MyType.
  function myArgFunction( x : MyType; y, z : integer ) return MyType;

  ---------------------------
  -- Component Declaration
  ---------------------------

  component ExampleComponent is
  generic (
    -- Board Specific Args
    BOARD_INFO    : BoardInfo;
    -- Streaming Input Interface Parameters
    DATA_IN_INFO  : AvalonSTInfo;
    -- Streaming Output Interface Parameters.
    DATA_OUT_INFO : AvalonSTInfo
  );
  port (
    -- Slave Interface
    slaveClk   : in  std_logic;
    slaveReset : in  std_logic;
    slaveIRQ   : out AvalonIRQSenderIntfArray(2 downto 0);
    slaveIn    : in  AvalonMMFabricToSlaveIntf;
    slaveInRsp : out AvalonMMSlaveToFabricIntf;
    -- Streaming Clock/Reset
    dataClk    : in  std_logic;
    dataReset  : in  std_logic;
    -- Streaming Input Interface
    dataIn     : in  AvalonSTSourceToSinkIntf;
    dataInRsp  : out AvalonSTSinkToSourceIntf;
    -- Streaming Output Interface
    dataOut    : out AvalonSTSourceToSinkIntf;
    dataOutRsp : in  AvalonSTSinkToSourceIntf
  );
  end component;

  ------------------------------------------------------------------------------
  -- ExampleSubSystem
  ------------------------------------------------------------------------------

  --! @brief ExampleSubSystem Configuration Parameters
  --! @details
  type ExampleSubSystemInfo is record
    --! @brief The number of DMA Engines of both types instantiated
    --! so NUM_DMA_ENGINES = 2 would result in 2 DMA_Readers and 2
    --! DMA Writers being instantiated
    NUM_DMA_ENGINES : natural range 1 to DMA_MAX_ENGINES/2;
  end record;

  --! @brief A default ExampleSubSystemInfo
  constant DEFAULT_EXAMPLE_SUB_SYSTEM_INFO : ExampleSubSystemInfo := ( NUM_DMA_ENGINES  => 1 );

  --! @brief Create an array type of the ExampleSubSystemInfo for dealing with multiples
  type ExampleSubSystemInfoArray is array (natural range <>) of ExampleSubSystemInfo;

  --! @brief Create an init function for easier maintenace if new parameters are ever added
  function init( num_dma : natural := DEFAULT_EXAMPLE_SUB_SYSTEM_INFO.NUM_DMA_ENGINES ) return ExampleSubSystemInfo;

  --! @brief Create a toArray function to easily construct arrays of the same info
  function toArray( info : ExampleSubSystemInfo := DEFAULT_EXAMPLE_SUB_SYSTEM_INFO; num : natural := 1 ) return ExampleSubSystemInfoArray;

  ---------------------------
  -- Component Declaration
  ---------------------------

  component ExampleSubSystem is
  generic (
    BOARD_INFO              : BoardInfo;
    SLAVE_OUT_INFO          : AvalonMMSlaveInfo;
    DATA_IN_INFO            : AvalonSTInfoArray;
    DATA_OUT_INFO           : AvalonSTInfoArray;
    EXAMPLE_SUB_SYSTEM_INFO : ExampleSubSystemInfo := DEFAULT_EXAMPLE_SUB_SYSTEM_INFO
  );
  port(
    -- Memory Mapped Slave Control/Status Interface
    slaveInClk    : in  std_logic;
    slaveInReset  : in  std_logic;
    slaveInIRQ    : out AvalonIRQSenderIntf;
    slaveIn       : in  AvalonMMFabricToSlaveIntf;
    slaveInRsp    : out AvalonMMSlaveToFabricIntf;
    -- Memory Mapped Slave Output Interface
    slaveOutClk   : in  std_logic;
    slaveOutReset : in  std_logic;
    slaveOut      : out AvalonMMFabricToSlaveIntf;
    slaveOutRsp   : in  AvalonMMSlaveToFabricIntf;
    -- Streaming Data Clock/Reset Interface
    dataClk       : in  std_logic;
    dataReset     : in  std_logic;
    -- Streaming Data Input Interface
    dataIn        : in  AvalonSTSourceToSinkIntfArray;
    dataInRsp     : out AvalonSTSinkToSourceIntfArray;
    -- Streaming Data Output Interface
    dataOut       : out AvalonSTSourceToSinkIntfArray;
    dataOutRsp    : in  AvalonSTSinkToSourceIntfArray
  );
  end component;

--------------------------------------------------------------------------------
end package;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- BRIEF DESCRIPTION:
--------------------------------------------------------------------------------
--! @brief
--! Package Body

--------------------------------------------------------------------------------
package body ExamplePkg is
--------------------------------------------------------------------------------

  ------------------------------------------------------------------------------
  ------------------------------------------------------------------------------
  function myFunction return MyType is
    variable r : MyType(0 to 1);
  begin
    -- Do stuff here...
    return r;
  end function;
  ------------------------------------------------------------------------------
  ------------------------------------------------------------------------------
  function myArgFunction( x : MyType; y, z : integer ) return MyType is
    variable r : MyType(7 downto 0);
  begin
    r := x;
    if( y <= z ) then
      -- Do some stuff...
    elsif( y > z ) then
      -- Do some other stuff...
    end if;
    return r;
  end function;
  ------------------------------------------------------------------------------
  ------------------------------------------------------------------------------
  function init( num_dma : natural := DEFAULT_EXAMPLE_SUB_SYSTEM_INFO.NUM_DMA_ENGINES )
                 return ExampleSubSystemInfo is
    variable r : ExampleSubSystemInfo;
  begin
    r.NUM_DMA_ENGINES := num_dma;
    return r;
  end function;
  ------------------------------------------------------------------------------
  ------------------------------------------------------------------------------
  function toArray( info : ExampleSubSystemInfo := DEFAULT_EXAMPLE_SUB_SYSTEM_INFO;
                    num : natural := 1 ) return ExampleSubSystemInfoArray is
    variable r : ExampleSubSystemInfoArray(0 to num-1);
  begin
    r := (others => info);
    return r;
  end function;
  ------------------------------------------------------------------------------
  ------------------------------------------------------------------------------

--------------------------------------------------------------------------------
end package body;
--------------------------------------------------------------------------------
