-- ************************************************************************** --
-- *************             BittWare Incorporated              ************* --
-- *************         9 Hills Ave, Concord, NH 03301         ************* --
-- ************************************************************************** --
-- LEGAL NOTICE:                                                              --
--                   Copyright (c) 2011 BittWare, Inc.                        --
--     The user is hereby granted a non-exclusive license to use and or       --
--       modify this code provided that it runs on BittWare hardware.         --
--     Usage of this code on non-BittWare hardware without the express        --
--        written permission of BittWare is strictly prohibited.              --
--                                                                            --
--  E-mail: support@bittware.com                    Tel: 603-226-0404         --
-- ************************************************************************** --
-- ABSTRACT:
-- ************************************************************************** --
-- DESIGN NOTES:
-- ************************************************************************** --

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.UtilityPkg.all;
use work.BoardInfoPkg.all;
use work.AvalonPkg.all;
use work.ExamplePkg.all;
use work.DMAPkg.all;
use work.MemoryMappedPkg.all;
use work.SubSystemPkg.all;

--------------------------------------------------------------------------------
-- ABSTRACT:
--------------------------------------------------------------------------------
--! @brief
--! The ExampleSubSystem is a component that simplifies using the StreamingToSlaveSubSystem
--! such that it is appropriate for the example projects.  It reduces the number
--! of options for simplicity.  The StreamingToSlaveSubSystem does the heavy lifting
--! and contains the blocks required to move data between Avalon-ST and Avalon-MM
--! interfaces.

--------------------------------------------------------------------------------
-- DESIGN DETAILS:
--------------------------------------------------------------------------------
--! @details
--! The component simple instantiates a StreamingToSlaveSubSystem with a limited
--! set of options

--------------------------------------------------------------------------------
-- NOTES:
--------------------------------------------------------------------------------
--! @note - All data inputs and outputs should be on the same clock domain.

--------------------------------------------------------------------------------
-- TODO:
--------------------------------------------------------------------------------
--! @todo None

--------------------------------------------------------------------------------
-- BUGS
--------------------------------------------------------------------------------
--! @bug None

--------------------------------------------------------------------------------
-- RETURN LINKS:
--------------------------------------------------------------------------------
--! <br> @ref ExampleSubSystem "Return to Top"

--------------------------------------------------------------------------------
entity ExampleSubSystem is
--------------------------------------------------------------------------------
generic (
  --! @brief Board Specific Parameters
  BOARD_INFO              : BoardInfo;
  --! @brief Slave Output Interface Parameters
  SLAVE_OUT_INFO          : AvalonMMSlaveInfo;
  --! @brief Streaming Input Interface Parameters
  DATA_IN_INFO            : AvalonSTInfoArray;
  --! @brief Streaming Output Interface Parameters
  DATA_OUT_INFO           : AvalonSTInfoArray;
  --! @brief Component Specific Parameters
  EXAMPLE_SUB_SYSTEM_INFO : ExampleSubSystemInfo
);
port(
  --! @name Memory Mapped Slave Control/Status Interface
  --! @{
  slaveInClk    : in  std_logic;
  slaveInReset  : in  std_logic;
  slaveInIRQ    : out AvalonIRQSenderIntf;
  slaveIn       : in  AvalonMMFabricToSlaveIntf;
  slaveInRsp    : out AvalonMMSlaveToFabricIntf;
  --! @}
  --! @name Memory Mapped Slave Output Data Interface
  --! @{
  slaveOutClk   : in  std_logic;
  slaveOutReset : in  std_logic;
  slaveOut      : out AvalonMMFabricToSlaveIntf;
  slaveOutRsp   : in  AvalonMMSlaveToFabricIntf;
  --! @}
  --! @name Streaming Data Clock/Reset Interface
  --! @{
  dataClk       : in  std_logic;
  dataReset     : in  std_logic;
  --! @}
  --! @name Streaming Data Input Interface
  --! @{
  dataIn        : in  AvalonSTSourceToSinkIntfArray;
  dataInRsp     : out AvalonSTSinkToSourceIntfArray;
  --! @}
  --! @name Streaming Data Output Interface
  --! @{
  dataOut       : out AvalonSTSourceToSinkIntfArray;
  dataOutRsp    : in  AvalonSTSinkToSourceIntfArray
  --! @}
);

--------------------------------------------------------------------------------
end entity;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- BRIEF LINKS:
--------------------------------------------------------------------------------
--! @brief
--! <b>Entity Description:</b> @ref ExampleSubSystem "Here..."<br>
--! <b>Address Map:</b> @ref ExampleSubSystem_ADDR_MAP "Here..."<br>
--! <b>Block Diagram:</b> @ref ExampleSubSystem_IMAGE "Here..."<br>
--! @details @anchor ExampleSubSystem_IMAGE

--------------------------------------------------------------------------------
-- BLOCK DIAGRAM:
--------------------------------------------------------------------------------
--! @details
--! @image html example/docs/images/example_sub_system.gif "Component Block Diagram"

--------------------------------------------------------------------------------
-- RETURN LINKS:
--------------------------------------------------------------------------------
--! <br> @ref ExampleSubSystem_BEH "Return to Top"

--------------------------------------------------------------------------------
architecture ExampleSubSystem_BEH of ExampleSubSystem is
--------------------------------------------------------------------------------

  --------------------------------------------
  -- Aliasing
  --------------------------------------------

  alias data_in      : AvalonSTSourceToSinkIntfArray(0 to dataIn'length-1) is dataIn;
  alias data_in_rsp  : AvalonSTSinkToSourceIntfArray(0 to dataInRsp'length-1) is dataInRsp;

  alias data_out     : AvalonSTSourceToSinkIntfArray(0 to dataOut'length-1) is dataOut;
  alias data_out_rsp : AvalonSTSinkToSourceIntfArray(0 to dataOutRsp'length-1) is dataOutRsp;

  --------------------------------------------
  -- Constants
  --------------------------------------------

  constant SLAVE_IN_INFO : AvalonMMSlaveInfoArray := toArray( init(16, 32) );

  constant STREAMING_TO_SLAVE_SUB_SYSTEM_INFO : StreamingToSlaveSubSystemInfo := ( INTERNAL_DMA_READERS => EXAMPLE_SUB_SYSTEM_INFO.NUM_DMA_ENGINES,
                                                                                   INTERNAL_DMA_WRITERS => EXAMPLE_SUB_SYSTEM_INFO.NUM_DMA_ENGINES,
                                                                                   DIRECT_DMA_READERS   => 0,
                                                                                   DIRECT_DMA_WRITERS   => 0,
                                                                                   DMA_CONTROLLER_INFO  => init(1024, 128, BOTH),
                                                                                   SLAVE_WINDOW_INFO    => init(init(false, 1, 0)) );

  --------------------------------------------
  -- Signals
  --------------------------------------------

  signal data_in_clk    : std_logic_vector(0 to dataIn'length-1);
  signal data_in_reset  : std_logic_vector(0 to dataIn'length-1);
  signal data_out_clk   : std_logic_vector(0 to dataOut'length-1);
  signal data_out_reset : std_logic_vector(0 to dataOut'length-1);

  ---------------------------
  -- Dump XML
  ---------------------------

  -- synthesis translate_off
  use work.ctype_h.all;
  use work.strings_h.all;
  use work.stdio_h.all;

  impure function dump_xml return boolean is
    constant NAME      : string := ExampleSubSystem'path_name;
    constant FILE_NAME : FileName := init( "project_config.xml" );
    variable xml_file  : CFILE;
  begin
    xml_file := fopen(FILE_NAME, "a");

    if( xml_file = 0 ) then
      report "ERROR: ExampleSubSystem.vhd - Cannot Open XML File " &  FILE_NAME
        severity failure;
    end if;

    fprintf(xml_file, "<ComponentInfo type=%s name=%s>\n", addQuotes("ExampleSubSystem"), addQuotes(NAME) );
    fprintf(xml_file, "%s", getXML("slave_out_info", SLAVE_OUT_INFO));
    for ii in DATA_IN_INFO'range loop
      fprintf(xml_file, "%s", getXML("data_in_info", ii, DATA_IN_INFO(ii)));
    end loop;
    for ii in DATA_OUT_INFO'range loop
      fprintf(xml_file, "%s", getXML("data_out_info", ii, DATA_OUT_INFO(ii)));
    end loop;
    fprintf(xml_file, "%s", getXML("example_sub_system_info.num_dma_engines", EXAMPLE_SUB_SYSTEM_INFO.NUM_DMA_ENGINES));
    fprintf(xml_file, "</ComponentInfo>\n");

    fclose(xml_file);

    return true;
  end function;

  constant DUMP_XML_FILES : boolean := dump_xml;
  -- synthesis translate_on

--------------------------------------------------------------------------------
begin
--------------------------------------------------------------------------------

  data_in_clk    <= (others => dataClk);
  data_in_reset  <= (others => dataReset);
  data_out_clk   <= (others => dataClk);
  data_out_reset <= (others => dataReset);

  streaming_to_slave_sub_system_map : StreamingToSlaveSubSystem
  generic map (
    BOARD_INFO                         => BOARD_INFO,
    SLAVE_IN_INFO                      => SLAVE_IN_INFO,
    SLAVE_OUT_INFO                     => SLAVE_OUT_INFO,
    DATA_IN_INFO                       => DATA_IN_INFO,
    DATA_OUT_INFO                      => DATA_OUT_INFO,
    STREAMING_TO_SLAVE_SUB_SYSTEM_INFO => STREAMING_TO_SLAVE_SUB_SYSTEM_INFO
  )
  port map (
    slaveInClk(0)   => slaveInClk,
    slaveInReset(0) => slaveInReset,
    slaveInIRQ      => slaveInIRQ,
    slaveIn(0)      => slaveIn,
    slaveInRsp(0)   => slaveInRsp,
    slaveOutClk     => slaveOutClk,
    slaveOutReset   => slaveOutReset,
    slaveOut        => slaveOut,
    slaveOutRsp     => slaveOutRsp,
    dataInClk       => data_in_clk,
    dataInReset     => data_in_reset,
    dataIn          => data_in,
    dataInRsp       => data_in_rsp,
    dataOutClk      => data_out_clk,
    dataOutReset    => data_out_reset,
    dataOut         => data_out,
    dataOutRsp      => data_out_rsp
  );

--------------------------------------------------------------------------------
end architecture;
--------------------------------------------------------------------------------
