-- ************************************************************************** --
-- *************             BittWare Incorporated              ************* --
-- *************         9 Hills Ave, Concord, NH 03301         ************* --
-- ************************************************************************** --
-- LEGAL NOTICE:                                                              --
--                   Copyright (c) 2012 BittWare, Inc.                        --
--     The user is hereby granted a non-exclusive license to use and or       --
--       modify this code provided that it runs on BittWare hardware.         --
--     Usage of this code on non-BittWare hardware without the express        --
--        written permission of BittWare is strictly prohibited.              --
--                                                                            --
--  E-mail: support@bittware.com                    Tel: 603-226-0404         --
-- ************************************************************************** --
-- ABSTRACT:
--   Example testbench for standard avalon based components.
-- ************************************************************************** --
-- DESIGN NOTES:
-- ************************************************************************** --

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.UtilityPkg.all;
use work.BoardInfoPkg.all;
use work.ComponentPkg.all;
use work.AvalonPkg.all;
use work.RegisterPkg.all;
use work.MemoryMappedPkg.all;

use work.ctype_h.all;
use work.strings_h.all;
use work.stdio_h.all;
use work.SimulationPkg.all;

use work.ExamplePkg.all;
use work.ProjectConfigPkg.all;

--------------------------------------------------------------------------------
entity tb_ExampleComponent is
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
end entity;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture tb_ExampleComponent_BEH of tb_ExampleComponent is
--------------------------------------------------------------------------------

  ---------------------------
  -- Testbench Configuration
  ---------------------------

  -- Define a local board info type for simulation
  -- This is required for components that have technology specific
  -- details that need to be configured.  Supported board info
  -- structures are defined in BoardInfoPkg.vhd
  -- ACTIVE_BOARD_INFO is set in the ProjectConfigPkg.vhd by afw.tcl
  constant BOARD_INFO : BoardInfo := ACTIVE_BOARD_INFO;

  -- Memory Interface Parameters
  constant MASTER_INFO : AvalonMMMasterInfo := ( PROFILE    => MASTER_PIPELINED,
                                                 ADDR_WIDTH => 12,
                                                 DATA_WIDTH => 8 );

  constant EXAMPLE_COMPONENT_ADDR_MAP : MMAddressMap := init( "example_component", 0, 0 );

  constant DECODER_ADDR_MAPS : MMAddressMapArray := (0 => EXAMPLE_COMPONENT_ADDR_MAP);
  -- OR
  --constant DECODER_ADDR_MAPS : MMAddressMapArray := toMMAddressMapArray( EXAMPLE_COMPONENT_ADDR_MAP );

  constant NUM_SLAVES : natural := DECODER_ADDR_MAPS'length;

  constant FABRIC_INFO : MMFabricInfo := ( FABRIC_TYPE  => STANDARD_PIPELINED_FABRIC,
                                           DECODE_START => MASTER_INFO.ADDR_WIDTH-4,
                                           ADDR_MAP     => getMMFabricAddressMap(DECODER_ADDR_MAPS) );

  -- Streaming Input Interface Parameters
  constant DATA_IN_INFO : AvalonSTInfo := ( PROFILE          => AVALON_ST_TYPICAL,
                                            BITS_PER_SYMBOL  => 8,
                                            SYMBOLS_PER_BEAT => 1,
                                            READY_LATENCY    => 0 );

  -- Streaming Output Interface Parameters
  constant DATA_OUT_INFO : AvalonSTInfo := ( PROFILE          => AVALON_ST_TYPICAL,
                                             BITS_PER_SYMBOL  => 8,
                                             SYMBOLS_PER_BEAT => 1,
                                             READY_LATENCY    => 0 );

  -- Setup the files
  constant SIMULATION_CONTROLLER_IN_FILE_INFO  : FileInfo := ( NAME   => resize("../inputs/ExampleComponent.cmd"),
                                                               FORMAT => HEX,
                                                               MODE   => READ );

  constant SIMULATION_CONTROLLER_LOG_FILE_INFO : FileInfo := ( NAME   => resize("tb.log"),
                                                               FORMAT => HEX,
                                                               MODE   => WRITE );

  constant DATA_IN_FILE_INFO  : FileInfo := ( NAME   => resize("../inputs/ExampleComponent_Input.txt"),
                                              FORMAT => HEX,
                                              MODE   => READ );

  constant DATA_OUT_FILE_INFO : FileInfo := ( NAME   => resize("ExampleComponent_Output.txt"),
                                              FORMAT => HEX,
                                              MODE   => WRITE );

  -- Simulation Parameters
  constant SIM_SETTLE_DELAY : time := 1 us;

  ---------------------------
  -- Signal Declarations
  ---------------------------

  signal power_on   : std_logic;
  signal end_of_sim : boolean;

  signal master_clk : std_logic;
  alias  slave_clk  : std_logic is master_clk;

  signal master_reset : std_logic;
  alias  slave_reset  : std_logic is master_reset;

  signal sim_ctrl_master_out     : AvalonMMMasterToFabricIntf;
  signal sim_ctrl_master_out_rsp : AvalonMMFabricToMasterIntf;
  signal sim_ctrl_end_of_file    : boolean;

  signal master_decoder_slave_out     : AvalonMMFabricToSlaveIntfArray(0 to NUM_SLAVES-1);
  signal master_decoder_slave_out_rsp : AvalonMMSlaveToFabricIntfArray(0 to NUM_SLAVES-1);

  signal data_clk   : std_logic;
  signal data_reset : std_logic;

  signal reader_data_out     : AvalonSTSourceToSinkIntf;
  signal reader_data_out_rsp : AvalonSTSinkToSourceIntf;
  signal reader_end_of_file  : boolean;

  signal example_component_data_out     : AvalonSTSourceToSinkIntf;
  signal example_component_data_out_rsp : AvalonSTSinkToSourceIntf;

--------------------------------------------------------------------------------
begin
--------------------------------------------------------------------------------

  ---------------------------
  -- TB Control Process
  ---------------------------

  ctrl_prc : process
  begin
    -- Turn on all clocks
    power_on <= '1';

    -- Clear sim flag
    end_of_sim <= false;

    -- Clock generators will come out of reset soon
    wait until (slave_reset = '0') and (data_reset = '0');

    -- Wait until all the ctrl/config is done
    printf("INFO: TB - Waiting on sim_ctrl_end_of_file.\n");
    wait until (sim_ctrl_end_of_file = true);

    -- Now wait for the data to pipe through the UUT
    printf("INFO: TB - Waiting on reader_end_of_file.\n");
    -- wait until doesn't trigger if the value is already true.
    -- it is really waiting an event and then it checks the value.
    -- Check the condition first so we aren't waiting forever...
    if( reader_end_of_file = false ) then
      wait until (reader_end_of_file = true);
    end if;

    -- After data has gone through the system, let all pipelines flush
    wait for SIM_SETTLE_DELAY;

    -- Make sure we signal endOfSim to close files...
    end_of_sim <= true;
    wait for 100 ns;

    -- Turn off all the clocks
    power_on <= '0';

    -- Force and end to the simulation.  May not be needed
    -- if your power_on shuts down all events...
    assert false report "NONE.  Simulation Complete!  Check log for pass/fail status." severity failure;

    -- We most likely won't get here, but this is the real way to end a sim...
    -- Sometimes however, it is very hard to shut down all events so
    -- a sim can end.  The assert above will end it though.  :)
    printf("INFO: TB - Last Wait!.\n");

    wait;
  end process;

  ---------------------------
  -- Generate Clocks
  ---------------------------

  clk_gen_master_map : ClkGen
  generic map (
    PERIOD       => 12 ns,
    RESET_CYCLES => 3,
    RESET_STATE  => '1'
  )
  port map (
    enable => power_on,
    clk    => master_clk,
    reset  => master_reset
  );

  clk_gen_data_map : ClkGen
  generic map (
    PERIOD       => 8 ns,
    RESET_CYCLES => 3,
    RESET_STATE  => '1'
  )
  port map (
    enable => power_on,
    clk    => data_clk,
    reset  => data_reset
  );

  ---------------------------
  -- Simulation Controller
  ---------------------------

  simulation_controller_map : SimulationController
  generic map (
    -- Simulation input and output file parameters
    IN_FILE_INFO  => SIMULATION_CONTROLLER_IN_FILE_INFO,
    OUT_FILE_INFO => SIMULATION_CONTROLLER_LOG_FILE_INFO,
    -- Avalon-MM interface parameters
    MASTER_INFO   => MASTER_INFO,
    -- User display data enable/disable
    VERBOSE       => true
  )
  port map (
    -- Avalon-MM Interface
    masterClk    => master_clk,
    masterReset  => master_reset,
    masterOut    => sim_ctrl_master_out,
    masterOutRsp => sim_ctrl_master_out_rsp,
    -- Simulation Flags
    endOfSim     => end_of_sim,
    endOfFile    => sim_ctrl_end_of_file
  );

  ---------------------------
  -- Master Decoder
  ---------------------------

  -- Simple direct connect functional fabric...
  --slave_in(0) <= toAvalonMMFabricToSlaveIntf( sim_ctrl_master_out );
  --sim_ctrl_master_out_rsp <= toAvalonMMFabricToMasterIntf( master_decoder_slave_out_rsp(0) );

  -- A larger fabric for multiple slave components
  master_decoder_map : MasterDecoder
  generic map (
    -- Fabric Parameters
    FABRIC_INFO => FABRIC_INFO,
    -- Avalon-MM interface parameters
    MASTER_INFO => MASTER_INFO
  )
  port map (
    -- Avalon-MM Master Interface
    masterClk   => master_clk,
    masterReset => master_reset,
    masterIn    => sim_ctrl_master_out,
    masterInRsp => sim_ctrl_master_out_rsp,
    -- Register Memory Interfaces
    slaveClk    => slave_clk,
    slaveReset  => slave_reset,
    slaveOut    => master_decoder_slave_out,
    slaveOutRsp => master_decoder_slave_out_rsp
  );

  ---------------------------
  -- Data Source
  ---------------------------

  avalon_st_file_reader_map : AvalonSTFileReader
  generic map (
    -- File Parameters
    FILE_INFO       => DATA_IN_FILE_INFO,
    -- Streaming Data Parameters
    DATA_INFO       => DATA_IN_INFO,
    -- Component Parameters
    SWAP_SYMBOLS    => false,
    SIGN            => false,
    SERIALIZE       => 1,
    BITS_PER_PACKET => 0
  )
  port map (
    -- Avalon-ST Interface
    dataClk    => data_clk,
    dataReset  => data_reset,
    dataOut    => reader_data_out,
    dataOutRsp => reader_data_out_rsp,
    -- Simulation Flags
    endOfSim   => end_of_sim,
    endOfFile  => reader_end_of_file
  );

  ---------------------------
  -- Component Under Test
  ---------------------------

  example_component_map : ExampleComponent
  generic map (
    -- Board Specific Args
    BOARD_INFO    => BOARD_INFO,
    -- Streaming Input Interface Parameters
    DATA_IN_INFO  => DATA_IN_INFO,
    -- Streaming Output Interface Parameters
    DATA_OUT_INFO => DATA_OUT_INFO
  )
  port map (
    -- Slave Interface
    slaveClk   => slave_clk,
    slaveReset => slave_reset,
    slaveIRQ   => open,
    slaveIn    => master_decoder_slave_out(EXAMPLE_COMPONENT_ADDR_MAP.MM_PORT),
    slaveInRsp => master_decoder_slave_out_rsp(EXAMPLE_COMPONENT_ADDR_MAP.MM_PORT),
    -- Streaming Clock/Reset
    dataClk    => data_clk,
    dataReset  => data_reset,
    -- Streaming Input Interface
    dataIn     => reader_data_out,
    dataInRsp  => reader_data_out_rsp,
    -- Streaming Output Interface
    dataOut    => example_component_data_out,
    dataOutRsp => example_component_data_out_rsp
  );

  ---------------------------
  -- Data Sink
  ---------------------------

  avalon_st_file_writer_map : AvalonSTFileWriter
  generic map (
    -- File Parameters
    FILE_INFO    => DATA_OUT_FILE_INFO,
    -- Streaming Data Parameters
    DATA_INFO    => DATA_OUT_INFO,
    -- Component Parameters
    SWAP_SYMBOLS => false,
    SIGN         => false,
    DESERIALIZE  => 1,
    BACKPRESSURE => true
  )
  port map (
    -- Avalon-ST Interface
    dataClk   => data_clk,
    dataReset => data_reset,
    dataIn    => example_component_data_out,
    dataInRsp => example_component_data_out_rsp,
    -- Simulation Flags
    endOfSim  => end_of_sim
  );

--------------------------------------------------------------------------------
end architecture;
--------------------------------------------------------------------------------
